//**************************************************************
//*            OpenGLide - Glide to OpenGL Wrapper
//*             http://openglide.sourceforge.net
//*
//*     SDL specific functions for handling display window
//*
//*         OpenGLide is OpenSource under LGPL license
//*              Originaly made by Fabio Barros
//*      Modified by Paul for Glidos (http://www.glidos.net)
//*               Linux version by Simon White
//**************************************************************
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef C_USE_SDL

#include <stdlib.h>
#include <math.h>

#include "GlOgl.h"

#include "platform/window.h"
#include "Glextensions.h"

static struct
{
    Uint16 red[ 256 ];
    Uint16 green[ 256 ];
    Uint16 blue[ 256 ];
} old_ramp;

static bool ramp_stored  = false;
static bool wasInit      = false;

void OpenGLRenderInit();
void OpenGLRenderShutdown();

bool InitialiseOpenGLWindow(FxU wnd, int x, int y, int width, int height)
{
    bool FullScreen = UserConfig.InitFullScreen;

	if ((width & 0x1) || (height & 0x1))
	{
		GlideMsg("Invalid resolution: %dx%d", width, height);
		return false;
	}

    wasInit = SDL_WasInit(SDL_INIT_VIDEO)!=0;
    if(!wasInit)
    {
        bool err = false;
        char *oldWindowId = 0;
        char windowId[40];

        if (wnd)
        {   // Set SDL window ID
            sprintf (windowId, "SDL_WINDOWID=%ld", (long)wnd);
            oldWindowId = SDL_getenv("SDL_WINDOWID");
            if (oldWindowId)
                oldWindowId = strdup(oldWindowId);
        }

        if (SDL_Init(SDL_INIT_VIDEO))
        {
            GlideMsg("Can't init SDL %s",SDL_GetError());
            err = true;
        }

        if (wnd)
        {   // Restore old value
            if (!oldWindowId)
                SDL_putenv("SDL_WINDOWID");
            else
            {
                sprintf (windowId, "SDL_WINDOWID=%s", oldWindowId);
                SDL_putenv(windowId);
                free (oldWindowId);
            }
        }

        if (err)
            return false;
    } else {
        SDL_Surface* tmpSurface = SDL_GetVideoSurface();
        if (tmpSurface)
        {
            // Preserve window/fullscreen mode in SDL apps and override config file entry
           (tmpSurface->flags&SDL_FULLSCREEN) ? (FullScreen = true) : (FullScreen = false);
        }
    }

	SDL_ShowCursor(SDL_DISABLE);

    SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	if (!UserConfig.StereoMode)
		SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);
	if (UserConfig.StereoMode == 1)
		SDL_GL_SetAttribute(SDL_GL_STEREO, 1);

	if((SDL_SetVideoMode(width, height, 32, FullScreen ? SDL_OPENGL|SDL_FULLSCREEN : SDL_OPENGL)) == 0)
    {
        GlideMsg("Video mode set failed: %s\n", SDL_GetError());
        return false;
    }

	if (!UserConfig.StereoMode)
	{
		SDL_GL_GetAttribute(SDL_GL_DEPTH_SIZE, &height);
		if ( height > 16 ) {
			UserConfig.PrecisionFix = false;
		}
	}

    if(SDL_GetGammaRamp(old_ramp.red, old_ramp.green, old_ramp.blue) != -1)
        ramp_stored = true;

#define GLPROC(pname,pparam) pname = (pparam)SDL_GL_GetProcAddress(#pname); \
	if (!pname) \
	{ \
		Error("Failed to get address of \""#pname"\""); \
		exit(1); \
	}
	
	GLPROC(glCreateShader,PFNGLCREATESHADERPROC)
	GLPROC(glDeleteShader,PFNGLDELETESHADERPROC)
	GLPROC(glShaderSource,PFNGLSHADERSOURCEPROC)
	GLPROC(glCompileShader,PFNGLCOMPILESHADERPROC)
	GLPROC(glGetShaderiv,PFNGLGETSHADERIVPROC)
	GLPROC(glGetShaderInfoLog,PFNGLGETSHADERINFOLOGPROC)
	GLPROC(glCreateProgram,PFNGLCREATEPROGRAMPROC)
	GLPROC(glDeleteProgram,PFNGLDELETEPROGRAMPROC)
	GLPROC(glAttachShader,PFNGLATTACHSHADERPROC)
	GLPROC(glLinkProgram,PFNGLLINKPROGRAMPROC)
	GLPROC(glUseProgram,PFNGLUSEPROGRAMPROC)
	GLPROC(glGetProgramiv,PFNGLGETPROGRAMIVPROC)
	GLPROC(glGetProgramInfoLog,PFNGLGETPROGRAMINFOLOGPROC)

	GLPROC(glGenBuffers,PFNGLGENBUFFERSPROC)
	GLPROC(glDeleteBuffers,PFNGLDELETEBUFFERSPROC)
	GLPROC(glBindBuffer,PFNGLBINDBUFFERPROC)
	GLPROC(glBufferData,PFNGLBUFFERDATAPROC)
	GLPROC(glMapBufferRange,PFNGLMAPBUFFERRANGEPROC)
	GLPROC(glUnmapBuffer,PFNGLUNMAPBUFFERPROC)

	GLPROC(glGenVertexArrays,PFNGLGENVERTEXARRAYSPROC)
	GLPROC(glBindVertexArray,PFNGLBINDVERTEXARRAYPROC)
	GLPROC(glDeleteVertexArrays,PFNGLDELETEVERTEXARRAYSPROC)
	GLPROC(glGetUniformLocation,PFNGLGETUNIFORMLOCATIONPROC)
	GLPROC(glUniform1i,PFNGLUNIFORM1IPROC)
	GLPROC(glUniformMatrix4fv,PFNGLUNIFORMMATRIX4FVPROC)
	GLPROC(glUniform4fv,PFNGLUNIFORM4FVPROC)
	GLPROC(glVertexAttribPointer,PFNGLVERTEXATTRIBPOINTERPROC)
	GLPROC(glEnableVertexAttribArray,PFNGLENABLEVERTEXATTRIBARRAYPROC)

	GLPROC(glGenFramebuffers,PFNGLGENFRAMEBUFFERSPROC)
	GLPROC(glBindFramebuffer,PFNGLBINDFRAMEBUFFERPROC)
	GLPROC(glDeleteFramebuffers,PFNGLDELETEFRAMEBUFFERSPROC)
	GLPROC(glCheckFramebufferStatus,PFNGLCHECKFRAMEBUFFERSTATUSPROC)
	GLPROC(glGenRenderbuffers,PFNGLGENRENDERBUFFERSPROC)
	GLPROC(glBindRenderbuffer,PFNGLBINDRENDERBUFFERPROC)
	GLPROC(glDeleteRenderbuffers,PFNGLDELETERENDERBUFFERSPROC)
	GLPROC(glRenderbufferStorage,PFNGLRENDERBUFFERSTORAGEPROC)
	GLPROC(glFramebufferRenderbuffer,PFNGLFRAMEBUFFERRENDERBUFFERPROC)
	GLPROC(glFramebufferTexture,PFNGLFRAMEBUFFERTEXTUREPROC)
	GLPROC(glDrawBuffers,PFNGLDRAWBUFFERSPROC)

	GLPROC(glActiveTexture,PFNGLACTIVETEXTUREPROC)

#undef GLPROC

	OpenGLRenderInit();

    return true;
}

void FinaliseOpenGLWindow( void)
{
	OpenGLRenderShutdown();

    if ( ramp_stored )
        SDL_SetGammaRamp(old_ramp.red, old_ramp.green, old_ramp.blue);
    if (!wasInit)
        SDL_Quit();
}

void SetGamma(float value)
{
    struct
    {
        Uint16 red[256];
        Uint16 green[256];
        Uint16 blue[256];
    } ramp;
    int i;

    for ( i = 0; i < 256; i++ )
    {
        Uint16 v = (Uint16)( 0xffff * pow( i / 255.0, 1.0 / value ) );

        ramp.red[ i ] = ramp.green[ i ] = ramp.blue[ i ] = ( v & 0xff00 );
    }

    SDL_SetGammaRamp(ramp.red, ramp.green, ramp.blue);
}

void RestoreGamma()
{
}

bool SetScreenMode(int &xsize, int &ysize)
{
    return true;
}

void ResetScreenMode()
{
}

void SwapBuffers()
{
    SDL_GL_SwapBuffers();
}

#endif // C_USE_SDL
